@@ -673,7 +673,7 @@ static void render_offsets(GtkHex *gh, gint imin, gint imax) {
 	gdk_gc_set_foreground(gh->offsets_gc, &GTK_WIDGET(gh)->style->text[GTK_STATE_NORMAL]);
 	
 	for(i = imin; i <= imax; i++) {
-		sprintf(offstr, "%08X", (gh->top_line + i)*gh->cpl);
+		sprintf(offstr, "%08X", (gh->top_line + i)*gh->cpl + gh->starting_offset);
 		/* Changes for Gnome 2.0 */
 		pango_layout_set_text (gh->olayout, offstr, 8);
 		gdk_draw_layout (w->window, gh->offsets_gc, 0, i*gh->char_height, gh->olayout);
@@ -1208,6 +1208,10 @@ static void gtk_hex_real_data_changed(GtkHex *gh, gpointer data) {
 
 	render_hex_lines(gh, start_line, end_line);
 	render_ascii_lines(gh, start_line, end_line);
+    if (gh->show_offsets)
+    {
+        render_offsets (gh, start_line, end_line);
+    }
 }
 
 static void bytes_changed(GtkHex *gh, gint start, gint end)
@@ -1223,6 +1227,10 @@ static void bytes_changed(GtkHex *gh, gint start, gint end)
 
 	render_hex_lines(gh, start_line, end_line);
 	render_ascii_lines(gh, start_line, end_line);
+    if (gh->show_offsets)
+    {
+        render_offsets (gh, start_line, end_line);
+    }
 }
 
 static void primary_get_cb(GtkClipboard *clipboard,
@@ -1598,7 +1606,7 @@ static void gtk_hex_finalize(GObject *o) {
 		(* G_OBJECT_CLASS(parent_class)->finalize)(G_OBJECT(o));  
 }
 
-static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
+static gboolean gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 	GtkHex *gh = GTK_HEX(w);
 	guint old_cp = gh->cursor_pos;
 	gint ret = TRUE;
@@ -1638,11 +1646,9 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 		}
 		break;
 	case GDK_Up:
-	case GDK_KP_8:
 		gtk_hex_set_cursor(gh, gh->cursor_pos - gh->cpl);
 		break;
 	case GDK_Down:
-	case GDK_KP_2:
 		gtk_hex_set_cursor(gh, gh->cursor_pos + gh->cpl);
 		break;
 	case GDK_Page_Up:
@@ -1659,7 +1665,6 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 		if(gh->active_view == VIEW_HEX)
 			switch(event->keyval) {
 			case GDK_Left:
-			case GDK_KP_4:
 				if(!(event->state & GDK_SHIFT_MASK)) {
 					gh->lower_nibble = !gh->lower_nibble;
 					if(gh->lower_nibble)
@@ -1670,7 +1675,6 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 				}
 				break;
 			case GDK_Right:
-			case GDK_KP_6:
 				if(gh->cursor_pos >= gh->document->file_size)
 					break;
 				if(!(event->state & GDK_SHIFT_MASK)) {
@@ -1715,6 +1719,16 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 					if(!gh->lower_nibble)
 						gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
 				}
+				else if((event->keyval >= GDK_KP_0)&&(event->keyval <= GDK_KP_9)) {
+					hex_document_set_nibble(gh->document, event->keyval - GDK_KP_0,
+											gh->cursor_pos, gh->lower_nibble,
+											gh->insert, TRUE);
+					if (gh->selecting)
+						gh->selecting = FALSE;
+					gh->lower_nibble = !gh->lower_nibble;
+					if(!gh->lower_nibble)
+						gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
+				}
 				else
 					ret = FALSE;
 
@@ -1723,11 +1737,9 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 		else if(gh->active_view == VIEW_ASCII)
 			switch(event->keyval) {
 			case GDK_Left:
-			case GDK_KP_4:
 				gtk_hex_set_cursor(gh, gh->cursor_pos - 1);
 				break;
 			case GDK_Right:
-			case GDK_KP_6:
 				gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
 				break;
 			default:
@@ -1741,6 +1753,14 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 					old_cp = gh->cursor_pos;
 					gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
 				}
+				else if((event->keyval >= GDK_KP_0)&&(event->keyval <= GDK_KP_9)) {
+					hex_document_set_byte(gh->document, event->keyval - GDK_KP_0 + '0',
+											gh->cursor_pos, gh->insert, TRUE);
+					if (gh->selecting)
+						gh->selecting = FALSE;
+					old_cp = gh->cursor_pos;
+					gtk_hex_set_cursor(gh, gh->cursor_pos + 1);
+				}
 				else
 					ret = FALSE;
 
@@ -1754,6 +1774,26 @@ static gint gtk_hex_key_press(GtkWidget *w, GdkEventKey *event) {
 	return ret;
 }
 
+static gboolean gtk_hex_key_release(GtkWidget *w, GdkEventKey *event) {
+	GtkHex *gh = GTK_HEX(w);
+
+	if(event->keyval == GDK_Shift_L || event->keyval == GDK_Shift_R) {
+		gh->selecting = FALSE;
+	}
+
+	return TRUE;
+}
+
+static gboolean gtk_hex_button_release(GtkWidget *w, GdkEventButton *event) {
+	GtkHex *gh = GTK_HEX(w);
+
+	if(event->state & GDK_SHIFT_MASK) {
+		gh->selecting = FALSE;
+	}
+
+	return TRUE;
+}
+
 /* 
  * recalculate the width of both displays and reposition and resize all
  * the children widgets and adjust the scrollbar after resizing
@@ -1815,6 +1855,13 @@ static gint gtk_hex_expose(GtkWidget *w, GdkEventExpose *event) {
 	return TRUE;
 }
 
+static void gtk_hex_document_changed(HexDocument* doc, gpointer change_data,
+        gboolean push_undo, gpointer data)
+{
+    gtk_hex_real_data_changed (GTK_HEX(data), change_data);
+}
+
+
 static void gtk_hex_size_request(GtkWidget *w, GtkRequisition *req) {
 	GtkHex *gh = GTK_HEX(w);
 	GtkRequisition sb_req;
@@ -1889,6 +1936,8 @@ static void gtk_hex_class_init(GtkHexClass *klass, gpointer data) {
 	GTK_WIDGET_CLASS(klass)->size_request = gtk_hex_size_request;
 	GTK_WIDGET_CLASS(klass)->expose_event = gtk_hex_expose;
 	GTK_WIDGET_CLASS(klass)->key_press_event = gtk_hex_key_press;
+	GTK_WIDGET_CLASS(klass)->key_release_event = gtk_hex_key_release;
+	GTK_WIDGET_CLASS(klass)->button_release_event = gtk_hex_button_release;
 	GTK_WIDGET_CLASS(klass)->realize = gtk_hex_realize;
 
 	/* Changed in Gnome 2.0 -- SnM */
@@ -1902,6 +1951,7 @@ static void gtk_hex_init(GtkHex *gh, gpointer klass) {
 
 	gh->disp_buffer = NULL;
 	gh->document = NULL;
+	gh->starting_offset = 0;
 
 	gh->xdisp_width = gh->adisp_width = 200;
 	gh->adisp_gc = gh->xdisp_gc = NULL;
@@ -2026,6 +2076,8 @@ GtkWidget *gtk_hex_new(HexDocument *owner) {
 	g_return_val_if_fail (gh != NULL, NULL);
 
 	gh->document = owner;
+    g_signal_connect (G_OBJECT (gh->document), "document_changed",
+            G_CALLBACK (gtk_hex_document_changed), gh);
 	
 	return GTK_WIDGET(gh);
 }
@@ -2073,6 +2125,8 @@ void gtk_hex_set_cursor(GtkHex *gh, gint index) {
 		if(!gh->insert && index == gh->document->file_size)
 			index--;
 
+		index == MAX(index, 0);
+
 		hide_cursor(gh);
 		
 		gh->cursor_pos = index;
@@ -2128,6 +2182,8 @@ void gtk_hex_set_cursor_xy(GtkHex *gh, gint x, gint y) {
 		if(!gh->insert && cp == gh->document->file_size)
 			cp--;
 
+		cp = MAX(cp, 0);
+
 		hide_cursor(gh);
 		
 		gh->cursor_pos = cp;
@@ -2199,7 +2255,7 @@ void gtk_hex_set_group_type(GtkHex *gh, guint gt) {
 /*
  * sets font for displaying data
  */
-void gtk_hex_set_font(GtkHex *gh, PangoFontMetrics *font_metrics, PangoFontDescription *font_desc) {
+void gtk_hex_set_font(GtkHex *gh, const PangoFontMetrics *font_metrics, const PangoFontDescription *font_desc) {
 	g_return_if_fail(gh != NULL);
 	g_return_if_fail(GTK_IS_HEX(gh));
 
@@ -2247,6 +2303,13 @@ void gtk_hex_show_offsets(GtkHex *gh, gboolean show)
 		hide_offsets_widget(gh);
 }
 
+void gtk_hex_set_starting_offset(GtkHex *gh, gint starting_offset)
+{
+	g_return_if_fail (gh != NULL);
+	g_return_if_fail(GTK_IS_HEX(gh));
+	gh->starting_offset = starting_offset;
+}
+
 void gtk_hex_set_insert_mode(GtkHex *gh, gboolean insert)
 {
 	g_return_if_fail(gh != NULL);
@@ -2290,7 +2353,7 @@ GtkHex_AutoHighlight *gtk_hex_insert_autohighlight(GtkHex *gh,
 												   gint len,
 												   const gchar *colour)
 {
-	GtkHex_AutoHighlight *new = g_malloc0(sizeof(GtkHex_Highlight));
+	GtkHex_AutoHighlight *new = g_malloc0(sizeof(GtkHex_AutoHighlight));
 
 	new->search_string = g_memdup(search, len);
 	new->search_len = len;
