@@ -110,6 +110,7 @@ struct _GtkHex
 	gint scroll_dir;
 	guint scroll_timeout;
 	gboolean show_offsets;
+	gint starting_offset;
 	gboolean insert;
 	gboolean selecting;
 };
@@ -140,8 +141,9 @@ guchar gtk_hex_get_byte(GtkHex *, guint);
 
 void gtk_hex_set_group_type(GtkHex *, guint);
 
+void gtk_hex_set_starting_offset(GtkHex *, gint);
 void gtk_hex_show_offsets(GtkHex *, gboolean);
-void gtk_hex_set_font(GtkHex *, PangoFontMetrics *, PangoFontDescription *);
+void gtk_hex_set_font(GtkHex *, const PangoFontMetrics *, const PangoFontDescription *);
 
 void gtk_hex_set_insert_mode(GtkHex *, gboolean);
 
@@ -167,6 +169,6 @@ GtkHex_AutoHighlight *gtk_hex_insert_autohighlight(GtkHex *gh,
                                                    const gchar *colour);
 void gtk_hex_delete_autohighlight(GtkHex *gh, GtkHex_AutoHighlight *ahl);
 
-G_BEGIN_DECLS
+G_END_DECLS
 
 #endif
